import time
import requests
import itertools
import json
import datetime 
import random



from IPython.core.display import display, HTML

def id_dict(queryList):
    newDict = {}
    for row in queryList:
        newDict[row['id']] = row
    return newDict

#login using username, password, and headers
#Should delete username and password as they are not used in the verification process
def open_rest_conn(headers,urlBase):
    # api url
    URL = urlBase + '/api-auth/login/'

    client = requests.session()

    # Retrieve the CSRF token first
    client.get(URL,headers=headers,verify=True)  # sets cookie
    if 'csrftoken' in client.cookies:
        # Django 1.6 and up
        csrftoken = client.cookies['csrftoken']
    else:
        # older versions
        csrftoken = client.cookies['csrf']

    login_data = dict(csrfmiddlewaretoken=csrftoken, next='/')
    temp_dict = dict(Referer=URL)
    client.post(URL, data=login_data, headers={**temp_dict,**headers})
    
    return client


def process_enc_stream(encRequest):
    superString = ""
    for t in encRequest:
        superString = superString + t.decode('utf8')

    #print(superString)

    stringDefs = [t.strip('(').strip(')') for t in superString.split("OrderedDict") if len(t)>0]

    myList=[]
    for i in range(len(stringDefs)):
        formatNamespace = {}        
        exec('myTuple = tuple('+stringDefs[i]+')', formatNamespace)
        myList.append(formatNamespace['myTuple'])

    return [{t[0][0]:t[0][1],t[1][0]:t[1][1]} for t in myList]   

#generator version of the above
def enc_generator(encRequest):
    leftoverString = ''
    for t in encRequest:
        stringList = (leftoverString+t.decode('utf8')).split('OrderedDict')
        #new value for the leftover string
        leftoverString = stringList[-1]
        #clean and format "good" strings from the split
        stringDefs = [u.strip('(').strip(')') for u in stringList[:-1] if len(u)>0]
        
        for i in range(len(stringDefs)):
            encNamespace = {}        
            exec('myTuple = tuple('+stringDefs[i]+')', encNamespace)
            tup = encNamespace['myTuple']
            yield {tup[0][0]:tup[0][1],tup[1][0]:tup[1][1]}
            
    encNamespace = {}        
    exec('myTuple = tuple('+leftoverString+')', encNamespace)
    tup = encNamespace['myTuple']
    yield {tup[0][0]:tup[0][1],tup[1][0]:tup[1][1]}

# an object intended to be in play everywhere managing info about the relation to the API
class CryptoContext:
    def __init__(self,headers,targetServer = False):
        if targetServer:
            print('Connecting to %s'%(targetServer,))
            self.urlBase = targetServer
        else:
            self.urlBase = 'https://ghostpii.com/api'
        
        self.client = open_rest_conn(headers,self.urlBase)

        self.headers = headers
        
        self.userInfo = self.client.get(
            self.urlBase+'/users/',
            headers = self.headers,
            verify = True
        ).json()
        
        self.userId = int([userDict['url'].split('/')[-2]  for userDict in self.userInfo][0])
        
        try:
            self.keyInfo = keyInfoPre.json()
        except:
            self.keyInfo = []
        
    def get(self,url,htmlDebug=False): 
    
        currentRequest =  self.client.get(
            self.urlBase + url,
            headers = self.headers,
            verify = True
        )
        if htmlDebug == True:
            
            return display(HTML(currentRequest.text))
        elif '/encrypt/' in url:
            
            return enc_generator(currentRequest)
        else:
            
            try:
                return currentRequest.json()
            except:
                print('json failure')
                print(display(HTML(currentRequest.text)))
            
    def post(self,url,myData):
        return self.client.post(
            self.urlBase + url,
            headers = self.headers,
            verify = True,
            data = myData
        )

#this returns a generator and not a completed object in memory
def encryption_key(apiContext,minMax,htmlDebug=False,seedString=False):
    if isinstance(seedString,str):
        print("Key generated locally from hash")
        return [{'id':t[0],'atom_key':t[1]} for t in zip(range(minMax[0],minMax[1]),string_to_key(seedString,minMax[1]-minMax[0]))]
    else:    
        url = '/staticencrypt/?lower=%d&upper=%d' % (minMax[0],minMax[1],)
        #print(url)
        return apiContext.get(url,htmlDebug)
    
    
def paillier_encryption_key(apiContext,htmlDebug=False,seedString=False):
    
    myKeyLoc = apiContext.get('/paillier-state/?length=1')
    #determine key boundaries
    dataBoundary = [myKeyLoc[0]['minId'],myKeyLoc[0]['maxId']]
    
    url = '/paillier-staticencrypt/?lower=%d&upper=%d' % (dataBoundary[0],dataBoundary[1])
    #print(url)
    return apiContext.get(url,htmlDebug)[0]

def decryption_key(apiContext,indicesJson,htmlDebug=False):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])

    #post a blob of information about the desired polynomial computation at the given timestamp
    
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":indicesJson,"userhash":timeStamp})
    #print(test.text)
    
    url = '/decrypt/?blobData=%d' % (
        timeStamp,
    )
    #print(url)
    return apiContext.get(url,htmlDebug)

def paillier_decryption_key(apiContext,index,htmlDebug=False):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])

    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/', {"assigned_user":apiContext.userId,"keyJSON":json.dumps([index]),"userhash":timeStamp,"enc_type":"Paillier"}
                          )
    #print(test)
    
    url = '/paillier-decrypt/?blobData=%d' % (
        timeStamp,
    )
    #print(url)
    return apiContext.get(url,htmlDebug)[0]

def linking_key(apiContext,indicesJson):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])

    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":indicesJson,"userhash":timeStamp})
    print(test)
    
    url = '/recordlink/?blobData=%d' % (
        timeStamp,
    )
    #print(url)    
    return apiContext.get(url)

def recryption_key(apiContext,requestData):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])

    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":indicesJson,"userhash":timeStamp})
    #print(test)
    
    url = '/recrypt/?blobData=%d' % (
        timeStamp,
    )
    #print(url)    
    return apiContext.get(url)

def ngram_checksum_key(apiContext,window,wordLength,indicesJson,isFloat=False):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])

    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":indicesJson,"userhash":timeStamp})
    #print(test)
    
    #this url
    url = '/ngramview/?win=%d&wordLength=%d&blobData=%d&isFloat=%s' % (
        window,
        wordLength,
        timeStamp,
        isFloat
    )
    return apiContext.get(url)

def polyn_comp_key(apiContext,polyn,polynVars,indicesTupleList,dualListOfList,isFloat=False,paillier=True):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])
    
    #information about the polynomial to be computed
    polynData = {
        'polyn':polyn,
        'polynVars':polynVars,
        'indicesTupleList':indicesTupleList,
        'dualTupleList':dualListOfList,
        'isFloat':isFloat,
        'paillier':paillier,
    }
    
    
    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":json.dumps(polynData),"userhash":timeStamp})
    #print(test.text)
    #this url tells the general polynomial endpoint to compute the desired polynomial
    url = '/general/?blobData=%d' % (
        timeStamp,
    )
    
    output = apiContext.get(url)
    #print(output)
    return output

def rand_poly_comp_key(apiContext,polyn,polynVars,indicesTupleList,dualListOfList,isFloat=False,paillier=True):
    
    #get the current timestamp
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])
    
    #information about the polynomial to be computed
    polynData = {
        'polyn':polyn,
        'polynVars':polynVars,
        'indicesTupleList':indicesTupleList,
        'dualTupleList':dualListOfList,
        'isFloat':isFloat,
        'paillier':paillier,
    }
    
    #print(polyn)
    
    #post a blob of information about the desired polynomial computation at the given timestamp
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":json.dumps(polynData),"userhash":timeStamp})
    #print(test.text)
    
    #this url tells the general polynomial endpoint to compute the desired polynomial
    url = '/randomized-poly/?blobData=%d' % (
        timeStamp,
    )
    
    output = apiContext.get(url)
    #print(output)
    return output

def paillier_convert(apiContext,paillierDict):
    
                    
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])
    #print(paillierDict)
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":json.dumps(paillierDict),"userhash":timeStamp})
    #print(test.text)
    
    url = '/paillier-otp/?blobData=%d' % (
        timeStamp,
    )
    
    newCipherData = apiContext.get(url)
    
    return newCipherData
    
def paillier_recrypt(apiContext,paillierDict):
    
                    
    timeStamp = int(str(datetime.datetime.now()).replace(' ','').replace('-','').replace(':','').replace('.','')[0:18])
    test = apiContext.post('/blob/',{"assigned_user":apiContext.userId,"keyJSON":json.dumps(paillierDict),"userhash":timeStamp})
    #print(test.text)
    
    url = '/paillier-recrypt/?blobData=%d' % (
        timeStamp,
    )
    
    newCipherData = apiContext.get(url)
    
    return newCipherData
    
    


from .ciphertext import *
from .db_toolbox import *
from .encoding import *
from .num_theory_toolbox import *
from .polynomial import *
from .seed_to_hash import *
from .recrypt import *

from .data_structures.norm_cipher_num import *
from .data_structures.norm_cipher_quant import *
from .data_structures.norm_cipher_string import *
from .data_structures.norm_cipher_list import *
from .data_structures.norm_cipher_frame import *
from .data_structures.paillier_num import *


from random import randint
import math
import json
from json import JSONEncoder
import numpy as np
import time
from ..db_toolbox import *

class PaillierInt():
    def __init__(self,pubKeyPair,cipher,fromPlain=False):
        if fromPlain == True:
            #in this scenario cipher is actually a plaintext integer
           
            self.n = pubKeyPair['n']
            self.n2 = pow(self.n,2)
            self.g = pubKeyPair['g']
            self.pubKeyPair = pubKeyPair
            
                     
                
            #final encryption step
            self.cipher = (pow(self.g,cipher,self.n2)*pow(randint(1,self.n),self.n,self.n2))%self.n2
            
                                      
        else:
                                      
            self.cipher = cipher
            self.n = pubKeyPair['n']
            self.g = pubKeyPair['g']
            self.n2 = pow(self.n,2)
            self.pubKeyPair = pubKeyPair
            
        
        
    def L(self,x):
        return (x-1)//self.n
    
    def __len__(self):
        return 1
        
    def __str__(self):
        return str(self.cipher)
    
    def __hash__(self):
        return sum([(t**2 % 10000) for t in cipherList]) % 10000
    
    def __add__(self,other):
        if type(other) is int:
            res = (self.cipher*pow(self.g,other,self.n2))%self.n2
            newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        elif type(other) is PaillierInt:
            res = (self.cipher*other.cipher) % self.n2
            newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        else:
            raise Exception("Can only add an int or another PaillierInt")
            
    def __radd__(self,other):
        if type(other) is int:
            res = (self.cipher*pow(self.g,other,self.n2))%self.n2
            newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        elif type(other) is PaillierInt:
            res = (self.cipher*other.cipher) % self.n2
            newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        else:
            raise Exception("Can only add an int or another PaillierInt")
            
            
    def __sub__(self,other):
        if type(other) is PaillierInt:
            res = (self.cipher * pow(other.cipher,-1,self.n2)) %self.n2
            newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        else:
            raise Exception("Can only subtract another PaillierInt")
            
        
    def __mul__(self,other):
        if type(other) is int:
            if other == 0:
                newNum = PaillierInt(self.pubKeyPair, 0,fromPlain=True)
            elif other == 1:
                newNum = PaillierInt(self.pubKeyPair, 0,fromPlain=True)+PaillierInt(self.pubKeyPair, self.cipher)
            else:
                res = pow(self.cipher,other,self.n2)
                newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        
    def __rmul__(self,other):
        if type(other) is int:
            if other == 0:
                newNum = PaillierInt(self.pubKeyPair, 0,fromPlain=True)
            elif other == 1:
                newNum = PaillierInt(self.pubKeyPair, 0,fromPlain=True)+PaillierInt(self.pubKeyPair, self.cipher)
            else:
                res = pow(self.cipher,other,self.n2)
                newNum = PaillierInt(pubKeyPair = self.pubKeyPair, cipher=res)
            return newNum
        
    def unsigned_decrypt(self,keyLambda,keyMu):
        
        plaintext = (self.L(pow(self.cipher,keyLambda,self.n2)) * keyMu) % self.n
        
        return plaintext
    
    def decrypt(self,keyLambda,keyMu):
        
        dec = self.unsigned_decrypt(keyLambda,keyMu)
        ans = ((dec + self.n//2)%self.n) - self.n//2
        return ans
    
    def ciphertext(self):
        
        return self.cipher
    
    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)
            
        
        
class PaillierFloat():
    def __init__(self,pubKeyPair,cipherTup,fromPlain=False,c=0,cIncrement = 20,r=False):
        self.large_num = 10**40
        if fromPlain:
            
            plain = cipherTup
            floor = math.floor(plain)
            res = plain-floor
            res = math.floor(res*self.large_num)
            self.n = pubKeyPair['n']
            self.n2 = pow(self.n,2)
            self.g = pubKeyPair['g']
            self.pubKeyPair = pubKeyPair
            
            plain_tup = (floor,res)
            
            self.cipherTup = self.encrypt_tuple(plain_tup,r)
            self.c = c
            self.cIncrement = cIncrement
            
            
        else:
            
            self.n = pubKeyPair['n']
            self.n2 = pow(self.n,2)
            self.g = pubKeyPair['g']
            self.pubKeyPair = pubKeyPair
            self.cipherTup = cipherTup
            self.c = c
            self.cIncrement = cIncrement
            
            
    def L(self,x):
        return (x-1)//self.n
    
    def encrypt_tuple(self,plain_tup,r=False):
        start_random = time.time()
        if r:
            r = r
        else:
            r = pow(randint(1,self.n),self.n,self.n2)
       
        new_g = pow(self.g,plain_tup[0],self.n2)
        enc_val1 = (new_g*r)%self.n2
        enc_val2 = (pow(self.g,plain_tup[1],self.n2)*r)%self.n2
        
        return (enc_val1,enc_val2)
    
    def unsigned_single_decrypt(self,single_val,keyLambda,keyMu):
        
        plaintext = (self.L(pow(single_val,keyLambda,self.n2)) * keyMu) % self.n
        
        return plaintext
    
    def single_decrypt(self,single_val,keyLambda,keyMu):
        
        dec = self.unsigned_single_decrypt(single_val,keyLambda,keyMu)
        ans = ((dec + self.n//2)%self.n) - self.n//2
        
        return ans
    
    def decrypt(self,keyLambda,keyMu):
        
        base = self.single_decrypt(self.cipherTup[0],keyLambda,keyMu)
        frac = self.single_decrypt(self.cipherTup[1],keyLambda,keyMu)
        
        return (base + frac/self.large_num)/(10**self.c)
    
    def recrypt(self,apiContext):
        tempOTP_int = random.randint(1,32767)
        tempOTP_frac = random.randint(1,32767)
        
        
        newNum = self + tempOTP_int 
        encryptedNums=[newNum.toJson()]
        isFloat = True
        
        paillierDict = {'keyID':[self.pubKeyPair['id']],'paillierData':encryptedNums,'isFloat':isFloat}
        newData = paillier_recrypt(apiContext,paillierDict)[0]
        #print(newData)
        
        self.cipherTup = (int(newData['cipherInt']),int(newData['cipherFrac']))
        
        self.cipherTup = (PaillierFloat(self.pubKeyPair,self.cipherTup)  - tempOTP_int).cipherTup
        self.c = 0
        return True
        
        
    
    def single_add(self, num1,num2):
        return (num1*num2)%self.n2
        
        
        
    # potential overflow with frac?
    # probably won't matter with our range but might be possible
    def __add__(self,other):
        if type(other) is PaillierFloat:
            if self.c > other.c:
                cdiff = self.c-other.c
                new_base = self.single_add(self.cipherTup[0],self.single_mult(other.cipherTup[0],10**cdiff))
                new_frac = self.single_add(self.cipherTup[1],self.single_mult(other.cipherTup[1],10**cdiff))
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
            elif other.c > self.c:
                cdiff = other.c-self.c
                new_base = self.single_add(self.single_mult(self.cipherTup[0],10**cdiff),other.cipherTup[0])
                new_frac = self.single_add(self.single_mult(self.cipherTup[1],10**cdiff),other.cipherTup[1])
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=other.c,cIncrement=self.cIncrement)
            else:
                new_base = self.single_add(self.cipherTup[0],other.cipherTup[0])
                new_frac = self.single_add(self.cipherTup[1],other.cipherTup[1])
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
        elif type(other) is int or type(other) is float:
            enc_other = PaillierFloat(self.pubKeyPair,other*(10**self.c),fromPlain=True,cIncrement=self.cIncrement)
            new_base = self.single_add(self.cipherTup[0],enc_other.cipherTup[0])
            new_frac = self.single_add(self.cipherTup[1],enc_other.cipherTup[1])
            return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
        else:
            raise Exception("Can only add an int, float, or another PaillierFloat")
            
    def __radd__(self,other):
        if type(other) is PaillierFloat:
            if self.c > other.c:
                cdiff = self.c-other.c
                new_base = self.single_add(self.cipherTup[0],self.single_mult(other.cipherTup[0],10**cdiff))
                new_frac = self.single_add(self.cipherTup[1],self.single_mult(other.cipherTup[1],10**cdiff))
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
            elif other.c > self.c:
                cdiff = other.c-self.c
                new_base = self.single_add(self.single_mult(self.cipherTup[0],10**cdiff),other.cipherTup[0])
                new_frac = self.single_add(self.single_mult(self.cipherTup[1],10**cdiff),other.cipherTup[1])
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=other.c,cIncrement=self.cIncrement)
            else:
                new_base = self.single_add(self.cipherTup[0],other.cipherTup[0])
                new_frac = self.single_add(self.cipherTup[1],other.cipherTup[1])
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
        elif type(other) is int or type(other) is float:
            enc_other = PaillierFloat(self.pubKeyPair,other*(10**self.c),fromPlain=True,cIncrement=self.cIncrement)
            new_base = self.single_add(self.cipherTup[0],enc_other.cipherTup[0])
            new_frac = self.single_add(self.cipherTup[1],enc_other.cipherTup[1])
            return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
        else:
            raise Exception("Can only add an int, float, or another PaillierFloat")
            
            
    def __sub__(self,other):
        
        neg_other = other * -1
        
        return PaillierFloat(self.pubKeyPair,self.cipherTup,c=self.c,cIncrement=self.cIncrement) + neg_other
            
        
    def single_mult(self,enc_num,scalar):
        
        res = pow(enc_num,scalar,self.n2)
            
        return res
        
        
    def __mul__(self,other):
        
        if type(other) is int:
            if other == 0:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)
            elif other == 1:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)+PaillierFloat(self.pubKeyPair,self.cipherTup,c=self.c,cIncrement=self.cIncrement)
            else:
                new_base = self.single_mult(self.cipherTup[0],other)
                new_frac = self.single_mult(self.cipherTup[1],other)
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
            
        elif type(other) is float or type(other) is np.float64:
            if other == 0:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)
            elif other == 1:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)+PaillierFloat(self.pubKeyPair,self.cipherTup,c=self.c,cIncrement=self.cIncrement)
            else:
                other_base = math.floor(other*(10**self.cIncrement))
                new_base = self.single_mult(self.cipherTup[0],other_base)
                new_frac = self.single_mult(self.cipherTup[1],other_base)
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c+self.cIncrement,cIncrement=self.cIncrement)
                
                
            
        else:
            print(other)
            raise Exception("only int and float multiplication are allowed with PaillierFloat")
        
    def __rmul__(self,other):
        
        if type(other) is int:
            if other == 0:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)
            elif other == 1:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)+PaillierFloat(self.pubKeyPair,self.cipherTup,c=self.c,cIncrement=self.cIncrement)
            else:
                new_base = self.single_mult(self.cipherTup[0],other)
                new_frac = self.single_mult(self.cipherTup[1],other)
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c,cIncrement=self.cIncrement)
            
        elif type(other) is float:
            if other == 0:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)
            elif other == 1:
                return PaillierFloat(self.pubKeyPair,0,fromPlain=True,cIncrement=self.cIncrement)+PaillierFloat(self.pubKeyPair,self.cipherTup,c=self.c,cIncrement=self.cIncrement)
            else:
                other_base = math.floor(other*(10**self.cIncrement))
                new_base = self.single_mult(self.cipherTup[0],other_base)
                new_frac = self.single_mult(self.cipherTup[1],other_base)
                return PaillierFloat(self.pubKeyPair,(new_base,new_frac),c=self.c+self.cIncrement,cIncrement=self.cIncrement)
                
                
            
        else:
            raise Exception("only int and float multiplication are allowed with PaillierFloat")
            
    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)
    
    



import math
import hashlib 

"""
this module provides functionality for deterministically producing an encryption key locally
the primary function is string_to_key 
this module is intended for use with the HashKey endpoint...
...the more compact seed string is transmitted to the API where it is used to generate key data

SEED STRINGS SHOULD BE DESTROYED AT EARLIEST POSSIBLE OPPORTUNITY
"""

# calculates the Shannon entropy of a string
def entropy(string):
        
        # get probability of chars in string
        prob = [ float(string.count(c)) / len(string) for c in dict.fromkeys(list(string)) ]

        # calculate the entropy
        entropy = - sum([ p * math.log(p) / math.log(2.0) for p in prob ])

        return entropy

# hash outputs are lists of characters, and this function matches trigrams to integers in the appropriate range    
def encoded_list_to_key(listOfHashes):
    keyList = []
    for i in range(len(listOfHashes)//3):
        keyAtom = 0
        for j in range(3):
            keyAtom+= (listOfHashes[3*i+j]*(102**j)) % 32767
        keyList.append(keyAtom)
    return keyList     

# this function enforces minimum standards on the complexity, in length and shannon entropy, of the seed string
def validate_seed(seedString):
    #input validation is very important here, need a string that is long and high entropy
    if len(seedString) > 8 and entropy(seedString) > 2:
        return True
    else:
        print("Please choose a longer and more complicated seed string.")
        return False

# this is the primary function intended for use outside this file
# this function takes a desired length and a complex seed string and produces an encryption key of appropriate length
# seed strings should not be retained in any form after submission to the API
def string_to_key(seedString,desiredLength):
    fullHashesRequired = desiredLength // 42
    lastHashLength = desiredLength - 42*fullHashesRequired
    
    superKeyList = []
    for i in range(fullHashesRequired):
        result = hashlib.sha512((seedString+str(i)).encode()) 
        superKeyList = superKeyList + encoded_list_to_key(list(result.hexdigest().encode()))
  
    #now compute keydata for the stub
    result = hashlib.sha512((seedString+str(fullHashesRequired+1)).encode()) 
    superKeyList = superKeyList + encoded_list_to_key(list(result.hexdigest().encode()))[0:lastHashLength]
    
    return superKeyList



import sympy as sym
import math
import random
import numpy as np
import json
import time

from src.ghostPii.data_structures.paillier_num import PaillierInt, PaillierFloat
from .db_toolbox import *
from .recrypt import paillier_to_otp

def create_polynomial(polyString,inVars):
    polyNamespace = {}
    exec("import sympy as sym", polyNamespace)    
    varList = []
    cList = []
    kList = []
    
    for varNameString in inVars:
        exec("c%s = sym.Symbol('c%s')" % (varNameString,varNameString,), polyNamespace)
        exec("k%s = sym.Symbol('k%s')" % (varNameString,varNameString,), polyNamespace)
        varList.append("c%s" % (varNameString,))
        varList.append("k%s" % (varNameString,))
        cList.append("c%s" % (varNameString,))
        kList.append("k%s" % (varNameString,))
        exec("%s = c%s - k%s" % (varNameString,varNameString,varNameString,), polyNamespace)  # plaintext integer
    
    exec('f='+polyString, polyNamespace)  # plaintext integer
    
    return {'namespace':polyNamespace,'varList':varList,'cList':cList,'kList':kList}

def flatten_nested(nestedTuple,length):
    endList = []
    remain = nestedTuple
    while len(endList)<length:
        new = []
        for t in remain:
            if isinstance(t,int):
                endList.append(t)
            else:
                new = new + list(t)
        remain = new
    return endList

from itertools import product
def all_partial_degrees(maxDegree,nVars):
    if nVars==1:
        for t in range(maxDegree+1):
            yield [t]
    else:
        baseProduct = range(maxDegree+1)
        nVars-=1
        for i in range(nVars):
            baseProduct = product(range(maxDegree+1),baseProduct)
        for t in baseProduct:
            flattenedT = flatten_nested(t,nVars)
            if(sum(flattenedT)<=maxDegree):
                yield flattenedT

def term_count(n,d):
    return int(sum([math.factorial(dd+n-1)/( math.factorial(dd) * math.factorial(n-1)) for dd in range(d+1)]))

def compute_polynomial(polynKey,indicesTupleList,cipherTupleList,polyn,paillier=True):
    #deserialize and reformat our key
    formattedKey = json.loads(polynKey[0]['keyJSON'])['monomialData']
    #list of client side ciphertext variables, need a count
    cipherVars = polyn['cList']
    #print(cipherVars)
    numberOfVars = len(cipherVars)
    #get the total degree of f
    try:
        f = polyn['namespace']['f']
        degree = sym.Poly(f).total_degree()
    except:
        degree = polyn['degree']
    if paillier:
        myPaillier = json.loads(polynKey[0]['keyJSON'])['paillierData']

        publicKey = {'n':myPaillier['n'],'g':myPaillier['g'],'id':myPaillier['id']}
        #print(publicKey)
        #temporary test Paillier computation

        formattedKeyEncrypted = {
            s:{t:PaillierFloat(publicKey,formattedKey[s][t],c=0) for t in formattedKey[s].keys()} 
            for s in formattedKey.keys()
        }
    else:
        #print(formattedKey)
        
        formattedKeyEncrypted = formattedKey
    
    values = []
    
    for fiber in zip(indicesTupleList,cipherTupleList):
        
        position=0
        currentIndices = fiber[0]
        currentCiphertext = fiber[1]
        currentKeyData = formattedKeyEncrypted[','.join([str(t) for t in currentIndices])]
        cipherMonomials = {}
        cipherMonomialStrings = []
        for exponentTuple in all_partial_degrees(degree,numberOfVars):
            #print('\n',exponentTuple,currentKeyData,currentCiphertext)
            monomial = 1
            cipherMonomialString = ''
            for i in range(len(exponentTuple)):
                exponent = exponentTuple[i]
                #construct a dictionary key for the current monomial
                currentVar = cipherVars[i]
                
                cipherMonomialString = cipherMonomialString + currentVar*exponent
                #take appropriate power of the derivative
                monomial = monomial*(currentCiphertext[i]**exponent)
            cipherMonomials[cipherMonomialString] = monomial
            cipherMonomialStrings.append(cipherMonomialString)
        if paillier:
            innerProductSum = PaillierFloat(publicKey,0,fromPlain=True,c=0)
            for monString in cipherMonomialStrings:
                currentProduct = currentKeyData[monString]*cipherMonomials[monString]

                innerProductSum = innerProductSum + currentProduct


            values.append(innerProductSum)
            
        else:
            
            values.append(sum([currentKeyData[monString]*cipherMonomials[monString] for monString in cipherMonomialStrings]))
        
    
    if paillier:
        return values,publicKey
    else:
        return values,None

def full_polynomial_compute(apiContext,polyString,variables,myIndices,myCiphers,isFloat,isSum=False,paillier=True,outPlain=False):
    
    if 'random' in polyString:
        myPoly = polyString
    else:
        myPoly = create_polynomial(polyString, variables)
    start = time.time()
    myKey = polyn_comp_key(apiContext,polyString,variables,myIndices,0,isFloat = isFloat,paillier=paillier)
    #print(myKey)
    end = time.time()
    
    paillier = json.loads(myKey[0]['keyJSON'])['paillier']
    #print(paillier)
    if 'random' in polyString:
        myPoly = json.loads(myKey[0]['keyJSON'])['poly']
        #print(myPoly)
    ans,paillierKey = compute_polynomial(myKey, myIndices, myCiphers, myPoly,paillier=paillier)
    end2 = time.time()
    
    
    if paillier:
        if isSum:
            ans = sum(ans)
            ncq = paillier_to_otp(apiContext,[ans],paillierKey['id'],isSum=isSum,isFloat=isFloat)
            end3 = time.time()
            
            if outPlain:
                return ncq.decrypt()[0]
            else:
                return ncq
        if outPlain:
            ncq = paillier_to_otp(apiContext,ans,paillierKey['id'],isSum=isSum,isFloat=isFloat)
            return ncq.decrypt()
            
        else:
            ncq = paillier_to_otp(apiContext,ans,paillierKey['id'],isSum=isSum,isFloat=isFloat)
            return ncq
    else:
        if outPlain:
            #print(ans)
            if isSum:
                return sum(ans)
            else:
                return ans
        else:
            from src.ghostPii.data_structures.norm_cipher_quant import NormCipherQuant
            if isSum:
                return NormCipherQuant(apiContext,[sum(ans)],fromPlain=True)
            else:
                return NormCipherQuant(apiContext,ans,fromPlain=True)
    end3 = time.time()
      
    return ncq
    
    


"""
I am currently assuming the n in my ngrams is 3
"""

import pickle
import numpy as np

# create an ngrams file and dump a pickle in the current working directory
def pickle_coding(fileName):
    if type(fileName)!=str:
        raise Exception("Filename must be a string")

    # this logarithm computation was used original to figure out what the "n" in ngram should be
    

    # these are the characters admitted in the ciphertext
    admChars = [chr(t) for t in range(33,123) if t not in [36,47,92]]
    # this forms the 3 fold combinations
    moreCharsEncode=[a+b+c for a in admChars for b in admChars for c in admChars]
    # open the file for writing
    fileObject = open(fileName,'wb') 
    # dump the list of ngrams as a file
    pickle.dump(moreCharsEncode,fileObject)   
    # close the file object
    fileObject.close()    

# this function takes a large integer index and returns a trigram representing it    
def proc_standard_encode(index):
    admChars = [chr(t) for t in range(33,123) if t not in [36,47,92]]
    interim = index
    trigram = ''
    for i in range(3):
        residue = interim % 87
        interim = (interim - residue) // 87
        trigram = trigram + admChars[residue]
    return trigram

# maps a python encoding index to its entry in the list of admissible characters
def adm_char_ind(index):
    if index < 36:
        return index - 33
    elif index < 47:
        return index - 34
    elif index < 92:
        return index - 35
    else:
        return index - 36

# this function takes a trigram and returns the index of the large integer it represents    
def proc_standard_decode(trigram):
    letterIndices = [adm_char_ind(t) for t in list(trigram.encode())]
    fullIndex = 0
    for i in range(3):
        fullIndex+=letterIndices[i]*(87**i)
    return fullIndex

# load the coding map of the givern filename pickled in the current working directory
# and return two functions wrapping an encoding list and a decoding dict, respectively
def load_coding(fileName):
    if type(fileName)!=str:
        raise Exception("Filename must be a string")

    try:
        ngramsFile = open(fileName,'rb')
        ngrams = pickle.load(ngramsFile)
        ngramsFile.close()  
    except:
        pickle_coding(fileName)
        ngramsFile = open(fileName,'rb')
        ngrams = pickle.load(ngramsFile)
        ngramsFile.close()  

    def encode_func(charInt):
        return ngrams[charInt]

    decodeDict = dict(zip(ngrams,range(len(ngrams))))
    def decode_func(trigram):
        return decodeDict[trigram]

    return {'encodeList':encode_func,'decodeDict':decode_func}

#second argument actually presumed to be a function wrapping a list
def encode_ciphertext(ciphertextIntegers,codingData):
    if codingData == False:
        encodeList = proc_standard_encode
    else:
        encodeList = codingData['encodeList']
        
    stringFragments = []
    for charInt in ciphertextIntegers:
        stringFragments.append(encodeList(charInt))
    return ''.join(stringFragments)

#second argument actually presumed to be a function wrapping a dictionary
def decode_ciphertext(myString,decodeDict):
    ciphertextIntegers = []
    for i in range(0, len(myString), 3):
        ciphertextIntegers.append(decodeDict(myString[i:(i+3)]))
    return ciphertextIntegers

#this function intended for a pandas frame of encrypted data encoded as ciphertext
def imp_dec_encrypted_frame(startingEncrypted,codingData):
    if codingData == False:
        decodeDict = proc_standard_decode
    else:
        decodeDict = codingData['decodeDict']


    goodIndices = [str(t) for t in range(0,-1+len(startingEncrypted.columns))]
    listOfList = [ [decode_ciphertext(str(s),decodeDict) for s in list(startingEncrypted[t])] for t in startingEncrypted[goodIndices]]
    colLengths = [ max([len(s) for s in colList]) for colList in listOfList ]

    return (listOfList,colLengths,len(startingEncrypted.index)) 


# basic modules
import numpy as np
import pandas as pd
import json
from sqlalchemy import *

#a tapas of additional scientific computing 
from scipy.spatial import distance

#capnion submodules

from ..encoding import *
from ..ciphertext import *
from ..num_theory_toolbox import *
from ..db_toolbox import *
from ..polynomial import *

import urllib.parse


#need to clarify how apiContexts should behave when these objects are added
#this is the main class for doing string computations on encrypted data
class NormCipherString:
    def __init__(self,apiContext,cipherList,indexData=False,keyRange=2000,permLevel='standard'):
        # api context is an important enduring property of the object
        self.apiContext = apiContext
        
        # if we are passed a string, we will encrypt it for the user
        if isinstance(cipherList,str):
            encoded_str = encode(cipherList)
            myLen = len(encoded_str)
            if permLevel == 'standard':
                myKeyLoc = apiContext.get('/state/?length='+str(myLen))[0]
            else:
                myKeyLoc = apiContext.get('/state/?length='+str(myLen)+'&permLevel='+urllib.parse.quote(permLevel)+'&keyRange='+str(keyRange))[0]
            #print(myKeyLoc)
            dataBoundary = [myKeyLoc['minId'],myKeyLoc['maxId']]
            myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=False)
            #print(myKeyGenerator)
            keyData = []
    
            for atom in myKeyGenerator:
                keyData.append(atom['atom_key'])
            encrypted_str = []
            for i in range(len(encoded_str)):
                encrypted_str.append(encoded_str[i]+keyData[i])
            
            indexData = dataBoundary[0]
            self.cipherList = encrypted_str
        else:
            self.cipherList = cipherList  
        self.length = len(cipherList)

        
        #index data may be an integer, indicating an idAnchor and continuous indices
        #or it may be a an explicit list of indices
        if isinstance(indexData,int):
            self.indicesList = [indexData+i for i in range(len(cipherList))]
        else:
            self.indicesList = indexData
            
        self.pairsList = zip(self.indicesList,self.cipherList)
        
    def __len__(self):
        return len(self.cipherList)
        
    def __str__(self):
        return ",".join(['('+str(t[0])+','+str(t[1])+')' for t in self.pairsList])
    
    def __hash__(self):
        return sum([t.residue() for t in self.cipherList]) % 32767 #max value for a key integer
    
    def __getitem__(self, key):
        return NormCipherString(
            self.apiContext,
            self.cipherList[key],
            indexData = self.indicesList[key]
        )
    
    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))
    
    def __iter__(self):
        for i in range(len(self.cipherList)):
            yield self[slice(i,i+1)]
    
    def __add__(self, other):
        return NormCipherString(
            self.apiContext,
            self.cipherList+other.cipherList,
            indexData = self.indicesList+other.indicesList
        )
    
    #this is key functionality and the most basic to require the API
    def __eq__(self, other): 
        if self.length != other.length:
            print("Ciphertext lengths are different.")
            return False
        else: 
            
            link = linking_key(self.apiContext,json.dumps(self.indicesList+other.indicesList))
            
            queryModulus = link[0]['prime']
            linkDict = id_dict(link)
            
            equal = False
            byCharEqual = []
            
            #we are using the presumption that the normcipherstrings are the same length
            for t in range(self.length):
                selfProd = polyn_comp_prod(linkDict[self.indicesList[t]],self.cipherList[t],queryModulus)
                otherProd = polyn_comp_prod(linkDict[other.indicesList[t]],other.cipherList[t],queryModulus)
                
                byCharEqual.append(selfProd == otherProd)
                
            equal = all(byCharEqual)
                
            return equal
        
    def __lt__(self,other):
        return [t.residue() for t in self.cipherList] < [t.residue() for t in other.cipherList]
    
    def pad(self,charsToAdd):
        encryptPadding = NormCipherString(self.apiContext,' '*charsToAdd)
        return NormCipherString(
            self.apiContext,
            self.cipherList + encryptPadding.cipherList,
            indexData = self.indicesList + encryptPadding.indicesList
        )
    
    #default to the standard encoding if no explicit map is given
    def ciphertext(self,encodeList=False):
        return encode_ciphertext(self.cipherList,encodeList)
    
    def decrypt(self):
        decryptKeyDict = {t['id']:t['atom_key'] for t in decryption_key(
            self.apiContext,
            json.dumps(self.indicesList)
        )}
        decryptKey = [decryptKeyDict[i] for i in self.indicesList]
        return ''.join([chr(t[0]-t[1]) for t in zip(self.cipherList,decryptKey)])      



#mostly a clone of the NCS class but with overloads that don't hit the API
#this is the secondary class for doing string computations on encrypted data
class AnalyticsCipherString:
    def __init__(self,apiContext,cipherList,helper,indexData,key):
        # api context is an important enduring property of the object
        self.apiContext = apiContext
        self.helper = helper
        self.num = key
        
        # if we are passed a string, we will encrypt it for the user
        if isinstance(cipherList,str):
            encoded_str = encode(cipherList)
            myLen = len(encoded_str)
            myKeyLoc = apiContext.get('https://ghostpii.com/state/?length='+str(myLen))[0]
            dataBoundary = [myKeyLoc['minId'],myKeyLoc['maxId']]
            encPerm = {'encrypt':dataBoundary}
            apiContext.post('https://ghostpii.com/keys/',{"assigned_user":apiContext.userId,"keyJSON":json.dumps(encPerm)} )
            myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=False)
            keyData = []
    
            for atom in myKeyGenerator:
                keyData.append(atom['atom_key'])
            encrypted_str = []
            for i in range(len(encoded_str)):
                encrypted_str.append(encoded_str[i]+keyData[i])
            
            indexData = dataBoundary[0]
            self.cipherList = encrypted_str
        else:
            self.cipherList = cipherList 
        self.length = len(cipherList)

        
        #index data may be an integer, indicating an idAnchor and continuous indices
        #or it may be a an explicit list of indices
        if isinstance(indexData,int):
            self.indicesList = [indexData+i for i in range(len(cipherList))]
        else:
            self.indicesList = indexData
            
        self.pairsList = zip(self.indicesList,self.cipherList)
        
    def __len__(self):
        return len(self.cipherList)
        
    def __str__(self):
        return ",".join(['('+str(t[0])+','+str(t[1])+')' for t in self.pairsList])
    
    def __hash__(self):
        return sum([t.residue() for t in self.cipherList]) % 32767 #max value for a key integer
    
    def __getitem__(self, key):
        if isinstance(key,slice):
            return AnalyticsCipherString(self.apiContext,[self.cipherList[key]],self.helper,self.indicesList[key], self.num+key.start)
        else:
            return AnalyticsCipherString(self.apiContext,[self.cipherList[key]],self.helper,self.indicesList[key], self.num+key)
    
    def __iter__(self):
        for i in range(len(self.cipherList)):
            yield self[slice(i,i+1)]
    
    #this is key functionality and the most basic to require the API
    def __eq__(self, other): 
        equals = self.helper.checkEquality(self.indicesList,other.indicesList,self.num,other.num)
        return equals
        
    #default to the standard encoding if no explicit map is given
    def ciphertext(self,encodeList=False):
        return encode_ciphertext(self.atomList,encodeList)

#strictly a helper function that stores some needed info about the NormCipherList
#and gives the AnalyticsCipherStrings access to that info for distances and equality checking
class AnalyticsHelper:
    
    def __init__(self,apiContext,matrix,indexOffset):
        self.apiContext = apiContext
        self.matrix = matrix
        self.offset = indexOffset
    
    def checkEquality(self,index1,index2,num1,num2):
        
        answer = True
        
        for i in range(len(index1)):
            ind1 = num1 + i
            ind2 = num2 + i
            if(self.matrix[ind1][ind2] == 1):
                answer = answer and True
            else:
                return False
            
        return answer




import random
import json

from .db_toolbox import *


def paillier_to_otp(apiContext,paillierEncrypted,paillierKeyID,isSum=False,isFloat=False):
    from .data_structures.norm_cipher_quant import NormCipherQuant
    from .data_structures.paillier_num import PaillierFloat


    length = len(paillierEncrypted)
    
    
    tempOTP_int = random.choices(list(range(1,32767)),k=length)
    
    #pad depending on float/int
    if isFloat:
        tempOTP_frac = random.choices(list(range(1,32767)),k=length)
        encryptedNums = []
        
        for i in range(length):
            newNum = paillierEncrypted[i] + (tempOTP_int[i] + tempOTP_frac[i]/32767.0)
            encryptedNums.append(json.dumps({'cipherTup':newNum.cipherTup, 'c':newNum.c}))
        
    else:
        
        encryptedNums = []
        for i in range(length):

            newNum = paillierEncrypted[i] + tempOTP_int[i]
            encryptedNums.append(json.dumps({'cipherTup':newNum.cipherTup, 'c':newNum.c}))
        
        
    paillierDict = {'keyID':[paillierKeyID],'paillierData':encryptedNums,'isFloat':isFloat,'isSum':isSum}
    #print(paillierDict)
    newCipherData = paillier_convert(apiContext,paillierDict)
    #print(newCipherData)
    
    cipherList = []
    indexList = []

    for i in range(len(newCipherData)):
        if isFloat:
            
            cipherList.append(newCipherData[i]['cipher']-(tempOTP_int[i]+tempOTP_frac[i]/32767.0))
        else:
            cipherList.append(int(newCipherData[i]['cipher']-tempOTP_int[i]))
        indexList.append(newCipherData[i]['keyID'])
    
    return NormCipherQuant(apiContext,cipherList,indexData=indexList,floatData=isFloat)






"""
This file contains functions for doing some basic number theory computations 
It also contains standard functions for pulling in stored number theory data like lists of prime numbers
"""

import time
import itertools
import numpy as np

def ext_gcd(aPre,bPre):
    try:
        ans = itersquare_long_cy.ext_gcd_cy(aPre,bPre)
        return {"coefOne":ans[0],"coefTwo":ans[1],"gcd":ans[2] }
    
    except:
        a = int(aPre)
        b = int(bPre)
        s = 0    
        old_s = 1
        t = 1    
        old_t = 0
        r = b
        old_r = a
        while r != 0:
            quotient = old_r // r
            old_r, r = r, old_r - quotient * r
            old_s, s = s, old_s - quotient * s
            old_t, t = t, old_t - quotient * t    
        return {"coefOne":old_s,"coefTwo":old_t,"gcd":old_r }  

# the basic modular arithmetic class
class ModArith:
    def __init__(self, residue, modulus):
        self.__residue = residue % modulus
        self.__modulus = modulus
    
    def __hash__(self):
        return self.__residue
        
    def __eq__(self, other): 
        if self.__modulus == other.__modulus:
            if self.__residue == other.__residue: 
                return True
            else:
                return False
        else:
            return "Moduli are not the same!"
        
    def __add__(self, other):
        if self.__modulus == other.__modulus:
            return ModArith((self.__residue + other.__residue) % self.__modulus,self.__modulus)
        else:
            return ModArith(0,1)
        
    def __mul__(self, other):
        if self.__modulus == other.__modulus:
            return ModArith((self.__residue * other.__residue) % self.__modulus,self.__modulus)
        else:
            return ModArith(0,1)
        
    def residue(self):
        return self.__residue
        
    def modulus(self):
        return self.__modulus        
        
    def twokth(self,k):
        prod = self
        
        for i in range(k):
            prod = prod*prod
            
        return prod
        
    def pow_fast(self,expo):
        #first try the C optimized code
        
        try :
            if self.residue() == 0:
                return ModArith(0,self.__modulus)
            else:
                return ModArith(itersquare_long_cy.c_pow_cy(self.__residue,expo,self.__modulus),self.__modulus)  
        #if C code doesn't work, revert to python
        except :
            result = ModArith(1,self.__modulus)
            x = self
            expoRun = expo 

            while expoRun > 0:
                if expoRun % 2 > 0:
                    result = result * x
                    expoRun = expoRun - 1
                x = x * x
                expoRun = expoRun / 2
            return result
    
    
    def inverse(self):
        euclData = ext_gcd(self.__residue,self.__modulus)
        if euclData['gcd']!=1:
            return ModArith(0,1)
        else:
            return ModArith(euclData['coefOne'],self.__modulus)
        
    def __str__(self):
        return "ModArith("+str(self.__residue)+","+str(self.__modulus)+")"

def pow_no_of(base,power,modulus):
    try:
        return itersquare_long_cy.pow_no_of_cy(base,power,modulus)
    except:
        result = base
        for i in range(power-1):
            result = (result * base) % modulus
        
        return result

# 0:4 ind_one,ind_two,row_base,row_prime,
# 4:7 square_coeff_one,lin_coeff_one,const_coeff_one
# 7:10 square_coeff_two,lin_coeff_two,const_coeff_two
def polyn_comp_prod(cursorRow,cipherInt,modulus):

    square = ModArith(cursorRow['square_coeff'],modulus).pow_fast(pow_no_of(cipherInt,2,(modulus-1)))
    linear = ModArith(cursorRow['lin_coeff'],modulus).pow_fast(cipherInt)
    constant = ModArith(cursorRow['const_coeff'],modulus)
    
    return square*linear*constant

#speed kills
#this is the good one
def compute_raw_links_c_poly_flat(ciphertextFlat,linkColumnCursor):
    #clock it
    start = time.time()
    # initialize the array to receive the matching data
    affinitiesFlat = []
    equivClass = []
    # use the cursor to iterate over the rows of the query
    for s in zip(ciphertextFlat,linkColumnCursor):
        equivClass.append(polyn_comp_prod(s[1],s[0],s[1]['prime']))
    for indTup in itertools.product(range(len(ciphertextFlat)),range(len(ciphertextFlat))): 
        #check equality of the computed comparison product with that from the query
        if equivClass[indTup[0]] == equivClass[indTup[1]]:
            affinitiesFlat.append(1)
        else:
            affinitiesFlat.append(0)
                
    outputArray = np.array(affinitiesFlat)
    outputArray.shape = (len(ciphertextFlat),len(ciphertextFlat))
    #print runtime
    print(time.time() - start)
            
    return outputArray




# basic modules
import numpy as np
import pandas as pd
import json
from sqlalchemy import *

#a tapas of additional scientific computing 
from scipy.spatial import distance

#capnion submodules

from ..encoding import *
from ..ciphertext import *
from ..num_theory_toolbox import *
from ..db_toolbox import *
from ..polynomial import *

class NormCipherNum:
    def __init__(self, apiContext,cipher,index=False,fromPlain=False,floatData=False,keyRange = 2000,permLevel='standard'):
        
        if fromPlain == True:
            #in this scenario cipher is actually a plaintext integer
            #register data
            myKeyLoc = apiContext.get('/state/?length=1'+'&range='+str(keyRange)+'&permLevel='+permLevel)
            
            #determine key boundaries
            dataBoundary = [myKeyLoc[0]['minId'],myKeyLoc[0]['maxId']]
            
           
            #pull enc key and encrypt
            myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=False)
            keyData = []
            if type(cipher) is int:
                for atom in myKeyGenerator:
                    keyData.append(atom['atom_key'])
                self.floatData = False
            else:
                for atom in myKeyGenerator:
                    keyData.append(atom['atom_key']+atom['atom_key_inv']/32767)
                self.floatData = True
                
            #final addition step
            self.cipher = cipher + keyData[0]
            self.index = dataBoundary[0]
            
                                      
        else:
                                      
            self.cipher = cipher
            self.index = index
            self.floatData = floatData
        
        self.apiContext = apiContext

        
    def __len__(self):
        return 1
        
    def __str__(self):
        return True
    
    def __hash__(self):
        return sum([(t**2 % 10000) for t in cipherList]) % 10000
    
    def __add__(self, other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        polyString = 'x+y'
        variables = ['x','y']
        
        ans = full_polynomial_compute(self.apiContext,polyString,variables,myIndices,
                                      cipherPair,self.floatData,paillier=False,outPlain=False)
        
        return ans[0]
        
    
    def __sub__(self, other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        polyString = 'x-y'
        variables = ['x','y']
        
        ans = full_polynomial_compute(self.apiContext,polyString,variables,myIndices,
                                      cipherPair,self.floatData,paillier=False,outPlain=False)
        
        return ans[0]
    
    
    def __mul__(self, other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        polyString = 'x*y'
        variables = ['x','y']
        
        ans = full_polynomial_compute(self.apiContext,polyString,variables,myIndices,
                                      cipherPair,self.floatData,paillier=False,outPlain=False)
        
        return ans[0]
    
    def __pow__(self, other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        polyString = 'x ** y'
        variables = ['x','y']
        
        ans = full_polynomial_compute(self.apiContext,polyString,variables,myIndices,
                                      cipherPair,self.floatData,paillier=False,outPlain=False)
        
        return ans[0]

    #this functions well but is not actually homomorphic
    def __lt__(self,other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return plainResults[0] < 0
       
    
    def __le__(self,other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return round(plainResults[0],2) <= 0
    
    def __gt__(self,other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return plainResults[0] > 0
    
    def __ge__(self,other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return round(plainResults[0],2) >= 0
        
    def __eq__(self, other): 
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return round(plainResults[0],2) == 0
    
    def __ne__(self, other):
        myIndices = [(self.index,other.index,),]
        cipherPair = [[self.cipher,other.cipher]]
        
        plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,
                                               isFloat=self.floatData,paillier=False,outPlain=True)
        return round(plainResults[0],2) != 0
    
    def decrypt(self):
    
        if type(self.cipher) is int:
            decryptKeyDict = {t['id']:t['atom_key'] for t in decryption_key(self.apiContext,json.dumps([self.index]))}
            decryptKey = decryptKeyDict[self.index]
            return self.cipher - decryptKey
        else:
            decryptKeyDict = {t['id']:(t['atom_key'],t['atom_key_inv']) for t in decryption_key(
                self.apiContext,
                json.dumps([self.index])
            )}
            decryptKey = decryptKeyDict[self.index]
            return self.cipher - (decryptKey[0]+decryptKey[1]/32767)
       
    
    def ciphertext(self,encodeList = False):
        
        return encode_ciphertext([self.cipher],encodeList)




# basic modules
import numpy as np
import pandas as pd
import json
from sqlalchemy import *
from copy import deepcopy

#a tapas of additional scientific computing 
from scipy.spatial import distance

#capnion submodules

from ..encoding import *
from ..ciphertext import *
from ..num_theory_toolbox import *
from ..db_toolbox import *
from ..polynomial import *

from .norm_cipher_list import *
from .norm_cipher_string import *
from .norm_cipher_quant import *


#this is a list of lists of data intended emulate a pandas data frame
class NormCipherFrame:

    def __init__(self,apiContext,cipherListOfListOfList,indexData=False,dataTypes=False,keyRange=2000,allFloat=False,permLevel='standard'):
        #get list of group elements
        #the rawest version of the underlying data is a list of list of integers
        self.apiContext = apiContext
        self.pure = False
        
        if isinstance(cipherListOfListOfList,pd.core.frame.DataFrame): #given a pandas data frame
            if not indexData: #presume the frame is plaintext
                
                self.dataTypes=[]
                for entry in cipherListOfListOfList.dtypes:
                    if entry == np.dtype('float64'):
                        self.dataTypes.append('float')
                    elif entry == np.dtype('int64') and not allFloat:
                        self.dataTypes.append('int')
                    elif entry == np.dtype('int64') and allFloat:
                        self.dataTypes.append('float')
                    elif entry == np.dtype('object'):
                        self.dataTypes.append('string')
                    else:
                        raise Exception('unknown data type')
                importData = import_and_encrypt2(
                    cipherListOfListOfList,
                    self.apiContext,
                    self.dataTypes,
                    seedString=False,
                    keyRange=keyRange,
                    allFloat=allFloat,
                    permLevel = permLevel
                )  
                self.cipherListOfListOfList = importData[0]
                indexData = importData[1]
                self.pure = True
                
            else: #presume the frame is ciphertext
                self.cipherListOfListOfList = imp_dec_encrypted_frame(cipherListOfListOfList,False)[0]
        elif isinstance(cipherListOfListOfList,list): #given a proper list of lists
            if not indexData: #assume the list is plaintext
                print('add, although this scenario does not come up much')
            else: #presume the list is ciphertext
                if isinstance(cipherListOfListOfList[0][0][0],int) or isinstance(cipherListOfListOfList[0][0][0],float): 
                    #directly given the encoded cipher integers
                    self.cipherListOfListOfList = cipherListOfListOfList
                    self.dataTypes = dataTypes
                else: #decode string data to obtain ciphertext integers
                    self.cipherListOfListOfList = [
                        [decode_ciphertext(cipher,proc_standard_decode) for cipher in listOfList] 
                        for listOfList in cipherListOfListOfList
                    ]
        else:
            print('data type not recognized')
                                         
        self.listOfColMaxChars = [len(listOfList[0]) for listOfList in self.cipherListOfListOfList] #assumes same length
        self.rows = len(self.cipherListOfListOfList[0])
        self.cols = len(self.cipherListOfListOfList) 
        self.lists = [i for i in range(self.cols)]
        
                                         
        if isinstance(indexData,int):
            self.indicesListOfListOfList = [] 
            i=indexData
            for cipherListOfList in self.cipherListOfListOfList:                             
                indicesListOfList = []                         
                for cipherList in cipherListOfList:
                    indicesListOfList.append(list(range(i,i+len(cipherList))))
                    i+=len(cipherList)
                self.indicesListOfListOfList.append(indicesListOfList)
        else:
            self.indicesListOfListOfList = indexData 
        
    def __len__(self):
        return self.cols

    def __getslice__(self, start, stop):
        return NormCipherFrame(
            self.apiContext,
            self.cipherListOfListOfList[slice(start,stop)],
            indexData = self.indicesListOfListOfList[slice(start,stop)],
            dataTypes = self.dataTypes[slice(start,stop)]
        )

    #approximating a list of strings, so getting an item should give the normcipherstring
    def __getitem__(self, key):
        if isinstance(key, slice):
            return NormCipherFrame(
                self.apiContext,
                self.cipherListOfListOfList[key],
                indexData = self.indicesListOfListOfList[key],
                dataTypes=self.dataTypes[key]
            )
        else:
            if isinstance(self.lists[key],int):
                if self.dataTypes[key] == 'string':
                    ncl = NormCipherList(
                        self.apiContext,
                        self.cipherListOfListOfList[key],
                        indexData = self.indicesListOfListOfList[key]
                        )
                    self.lists[key] = ncl
                    return ncl
                elif self.dataTypes[key] == 'int' :
                    ncq = NormCipherQuant(
                        self.apiContext,
                        [self.cipherListOfListOfList[key][i][0] for i in range(self.rows)],
                        indexData = [self.indicesListOfListOfList[key][i][0] for i in range(self.rows)],
                        floatData = False
                    )
                    self.lists[key]=ncq
                    return ncq
                else:
                    ncq = NormCipherQuant(
                        self.apiContext,
                        [self.cipherListOfListOfList[key][i][0] for i in range(self.rows)],
                        indexData = [self.indicesListOfListOfList[key][i][0] for i in range(self.rows)],
                        floatData = True
                    )
                    self.lists[key]=ncq
                    return ncq
            else:
                return self.lists[key]
        
    def __iter__(self):
        for i in range(len(self.cipherListOfListOfList)):
            yield self[i]
            
    def group_by(self,colNum):
        listOfGroupedNCF = []
        indicesAlreadyAdded = []
        colToSplit = self[colNum]

        temp_mx = self[colNum].char_equal_mx()
        colToSplit.helper = AnalyticsHelper(self.apiContext,temp_mx,colToSplit.colMaxChars)

        checksumList = colToSplit.checksum()
        for i in range(len(colToSplit)):
            newCipherFrame = None
            value = checksumList[i]
            newCipherListOfList = []
            newIndicesListOfList = []
            if i not in indicesAlreadyAdded:
                indicesAlreadyAdded.append(i)
                
                for j in range(len(colToSplit)):
                    ncs1 = colToSplit[i]
                    ncs2 = colToSplit[j]
                    if colToSplit.helper.checkEquality(ncs1.indicesList,
                                                       ncs2.indicesList,
                                                       i*colToSplit.colMaxChars,
                                                       j*colToSplit.colMaxChars):
                        if newCipherFrame is not None:
                            newCipherFrame = newCipherFrame.vert_merge(self.vert_slice([j]))
                        else:
                            newCipherFrame = self.vert_slice([j])
                        indicesAlreadyAdded.append(j)
                listOfGroupedNCF.append(newCipherFrame)

        return listOfGroupedNCF
            
    def horiz_merge(self, other):
        if isinstance(other, NormCipherFrame):
            return NormCipherFrame(
                self.apiContext,
                self.cipherListOfListOfList+other.cipherListOfListOfList,
                indexData = self.indicesListOfListOfList+other.indicesListOfListOfList,
                dataTypes = self.dataTypes + other.dataTypes
            )
        elif isinstance(other,NormCipherList):
            newCipherList = deepcopy(self.cipherListOfListOfList)
            newCipherList.append(other.cipherListOfList)
            newIndicesList = deepcopy(self.indicesListOfListOfList)
            newIndicesList.append(other.indicesListOfList)
            newDataTypes = deepcopy(self.dataTypes)
            newDataTypes.append('string')
            return NormCipherFrame(
                    self.apiContext,
                    newCipherList,
                    indexData = newIndicesList,
                    dataTypes = newDataTypes
                    )
        else:
            newCipherList = deepcopy(self.cipherListOfListOfList)
            newCipherList.append([[num] for num in other.cipherList])
            newIndicesList = deepcopy(self.indicesListOfListOfList)
            newIndicesList.append([[index]for index in other.indicesList])
            newDataTypes = deepcopy(self.dataTypes)
            if other.floatData:
                newDataTypes.append('float')
            else:
                newDataTypes.append('int')
            return NormCipherFrame(
                    self.apiContext,
                    newCipherList,
                    indexData = newIndicesList,
                    dataTypes = newDataTypes
                    )
            
    
    def vert_merge(self,other):
        return NormCipherFrame(
            self.apiContext,
            [
                self.cipherListOfListOfList[i]+other.cipherListOfListOfList[i] 
                for i in range(len(self.cipherListOfListOfList))
            ],
            indexData = [
                self.indicesListOfListOfList[i]+other.indicesListOfListOfList[i]
                for i in range(len(self.indicesListOfListOfList))
            ],
            dataTypes = self.dataTypes
        )        
    
    def vert_slice(self,rowList):
        return NormCipherFrame(
            self.apiContext,
            [
                [cipherListOfList[i] for i in rowList]
                for cipherListOfList in self.cipherListOfListOfList
            ],
            indexData = [
                [indicesListOfList[i] for i in rowList]
                for indicesListOfList in self.indicesListOfListOfList
            ],
            dataTypes=self.dataTypes
        )
    
    def merge(self,other,colIndexTuple,how='inner'):
        #get the matches on the original data
        myMatches = self[colIndexTuple[0]].generate_matches(other[colIndexTuple[1]])
        if how=='inner':
            #return appropriate entries in a horizontally merged NCF
            return self.vert_slice([t[0] for t in myMatches]).horiz_merge(
                other.vert_slice([t[1] for t in myMatches])
            )
        if how=='left':
            #right frame with an added empty row
            nullOther = other.vert_merge(NormCipherFrame(other.apiContext,pd.DataFrame([[' '*i for i in other.listOfColMaxChars]])))
            #what are the unmatched entries on the left
            missing = [(i,nullOther.rows-1) for i in range(self.rows) if i not in [tup[0] for tup in myMatches]]
            #add matches for unmatched rows onto the final empty row
            myMatches = myMatches + missing
            myMatches.sort(key=lambda t:t[0])
            #return appropriate entries in a horizontally merged NCF
            return self.vert_slice([t[0] for t in myMatches]).horiz_merge(
                nullOther.vert_slice([t[1] for t in myMatches])
            )
        if how=='right':
            #left frame with an added empty row
            nullSelf = self.vert_merge(NormCipherFrame(self.apiContext,pd.DataFrame([[' '*i for i in self.listOfColMaxChars]])))
            #what are the unmatched entries on the right
            missing = [(nullSelf.rows-1,i) for i in range(other.rows) if i not in [tup[1] for tup in myMatches]]
            #add matches for unmatched rows onto the final empty row
            myMatches = myMatches + missing
            myMatches.sort(key=lambda t:t[1])
            #return appropriate entries in a horizontally merged NCF
            return nullSelf.vert_slice([t[0] for t in myMatches]).horiz_merge(
                other.vert_slice([t[1] for t in myMatches])
            )
        if how=='outer':
            #left frame with an added empty row
            nullSelf = self.vert_merge(NormCipherFrame(self.apiContext,pd.DataFrame([[' '*i for i in self.listOfColMaxChars]])))
            #right frame with an added empty row
            nullOther = other.vert_merge(NormCipherFrame(other.apiContext,pd.DataFrame([[' '*i for i in other.listOfColMaxChars]])))
            #what are the unmatched entries on the left
            missingSelf = [(i,nullSelf.rows-1) for i in range(nullOther.rows) if i not in [tup[0] for tup in myMatches]][:-1]
            #what are the unmatched entries on the right
            missingOther = [(nullOther.rows-1,i) for i in range(nullSelf.rows) if i not in [tup[1] for tup in myMatches]][:-1]
            #add matches for unmatched rows onto the final empty row
            myMatches = myMatches + missingSelf + missingOther
            myMatches.sort(key=lambda t:t[0])
            #return appropriate entries in a horizontally merged NCF
            return nullSelf.vert_slice([t[0] for t in myMatches]).horiz_merge(
                nullOther.vert_slice([t[1] for t in myMatches])
            )
                    
    def lol_of_ciphertext(self,codingData=False):        
        return [self[t].list_of_ciphertext(codingData) for t in range(len(self))]
    
    def frame_of_ciphertext(self,codingData=False):   
        return pd.DataFrame(
            np.transpose(np.array(self.lol_of_ciphertext(codingData))),
            columns=list(range(len(self)))
        )
    
    def metadata(self):
        if self.pure:
            return json.dumps({
                'pure':self.pure,
                'listOfColMaxChars':self.listOfColMaxChars,
                'rows':self.rows,
                'cols':self.cols,
                'bounds':self.indicesListOfListOfList[0][0][0]
            })        
        else:
            return json.dumps({
                'pure':self.pure,
                'listOfColMaxChars':self.listOfColMaxChars,
                'rows':self.rows,
                'cols':self.cols,
                'bounds':(
                    self.indicesListOfListOfList[0][0][0],
                    self.indicesListOfListOfList[0][0][0]+sum(self.listOfColMaxChars)*self.rows
                )
            })
    
    def decrypt(self):
        decryptKeyDict = {t['id']:(t['atom_key'],t['atom_key_inv']) for t in decryption_key(
            self.apiContext,
            json.dumps(flatten_list(flatten_list(self.indicesListOfListOfList)))
        )}
        decryptKey = [decryptKeyDict[i] for i in flatten_list(flatten_list(self.indicesListOfListOfList))]
        i=0
        plainBig = []
        for k in range(len(self.cipherListOfListOfList)):
            encryptedWordList = self.cipherListOfListOfList[k]
            plain = []
            for j in range(len(encryptedWordList)):
                if self.dataTypes[k]=='float':
                    plain.append(encryptedWordList[j][0]-(decryptKey[i][0]+decryptKey[i][1]/32767.0))
                    i+=1
                elif self.dataTypes[k]=='int':
                    plain.append(encryptedWordList[j][0]-decryptKey[i][0])
                    i+=1
                else:        
                    encryptedWord = encryptedWordList[j]
                    plain.append(''.join([chr(t[0]-t[1][0]) for t in zip(encryptedWord,decryptKey[i:(i+len(encryptedWord))])]))
                    i+=len(encryptedWord)
            plainBig.append(plain)
        return plainBig
        
    def to_sql(self,table_name,connection):
        cipherFrame = self.frame_of_ciphertext()
        indexData = self.indicesListOfListOfList
        for column in range(len(indexData)):
            ids_to_append = []
            for entry in indexData[column]:
                ids_to_append.append(entry[0])
            cipherFrame['id{}'.format(column)] = ids_to_append
    
        return cipherFrame.to_sql(table_name,connection)
    
 
    
def import_and_encrypt2(myPlaintext,apiContext,dataTypes,desiredPerms=False,seedString=False,keyRange=32766,allFloat=False,permLevel='standard'):
    
    colIndex = 0
    cipherListOfListOfList = []
    indexListOfListOfList = []
    for (columnName, columnData) in myPlaintext.iteritems():
        #pad, measure, and encode
        if dataTypes[colIndex] == 'string':
            myEncodedList = encode_list(columnData)
            myLen = len(myEncodedList)*len(myEncodedList[0])
        elif dataTypes[colIndex] =='float':
            myEncodedList = [[float(f)] for f in columnData]
            myLen = len(myEncodedList) 
        else:
            myEncodedList = [[int(f)] for f in columnData]
            myLen = len(myEncodedList)
        #print(myEncodedList)
        #compute length of required encryption key
        
        
        #register key data of required length
        
        if isinstance(seedString,str):
            myKeyLoc = apiContext.get('/statehash/?length=%d&seedString=%s'%(myLen,seedString,))[0]
            #print('basic state')
        else:
            myKeyLoc = apiContext.get('/state/?length='+str(myLen)+'&range='+str(keyRange)+'&permLevel='+permLevel)[0]

        
        #create encryption key for current user
        
        dataBoundary = [myKeyLoc['minId'],myKeyLoc['maxId']]

        
        #pull enc key and encrypt

        myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=seedString)
        #print(myKeyGenerator)
        sortedKeyGenerator = {}
        for atom in myKeyGenerator:
            sortedKeyGenerator[str(atom['id'])] = atom
        #print(sortedKeyGenerator)
        currentIndex = myKeyLoc['minId']
        cipherListOfList = []
        indexListOfList = []
        for i in range(len(myEncodedList)):
            cipherList = []
            indexList = []
            for j in range(len(myEncodedList[0])):
                if dataTypes[colIndex] == 'float':
                    
                    cipherList.append(myEncodedList[i][j]+int(sortedKeyGenerator[str(currentIndex)]['atom_key'])+int(sortedKeyGenerator[str(currentIndex)]['atom_key_inv'])/32767.0)
                else:
                    cipherList.append(myEncodedList[i][j]+int(sortedKeyGenerator[str(currentIndex)]['atom_key']))
                indexList.append(currentIndex)
                currentIndex += 1
            cipherListOfList.append(cipherList)
            indexListOfList.append(indexList)        
            
        
        cipherListOfListOfList.append(cipherListOfList)
        indexListOfListOfList.append(indexListOfList)
        
        colIndex += 1


    return (
        cipherListOfListOfList,
        indexListOfListOfList,
    )
    
def import_from_db(cryptoContext,table_name,connection):
    meta_df = pd.read_sql_table(table_name,connection)
    meta_df = meta_df.drop('index',axis=1)
    num_rows = meta_df.shape[0]
    num_cols = meta_df.shape[1]//2
    indices = np.zeros((num_cols,num_rows),dtype = int)
    for index,row in meta_df.iterrows():
        for i in range(num_cols):
            indices[i,index] = row['id{}'.format(i)]
    indices_list = indices.tolist()
    for i in range(indices.shape[0]):
        col_length = len(meta_df[str(i)][0])//3
        for j in range(indices.shape[1]):
            indices_list[i][j] = [k for k in range(indices[i][j],indices[i][j]+col_length)]
        meta_df = meta_df.drop('id{}'.format(i),axis=1)
    return NormCipherFrame(cryptoContext,meta_df,indexData=indices_list)


import copy

def systematic_padding(listStrings,**kwargs):
    
    #examine different options for handling field length
    if 'maxLength' in kwargs:
        maxLength = kwargs.get('maxLength')
        keyLength = len(listStrings)*maxLength
    else:
        maxLength = max([len(t) for t in listStrings])
        keyLength = len(listStrings)*maxLength
    
    #create a new list if need be with all fields of the desired length
    return [t.ljust(maxLength) for t in listStrings]

def flatten_list(listOfList):
    return [item for sublist in listOfList for item in sublist]

# wrapper for the standard encoding
def encode(myString):
    return list(myString.encode())

# take a pandas frame, coerce to strings, pull information on length, pad to uniform length, and coerce to list of lists
def string_frame(dataFrame):
    listOfList = [ [str(s) for s in list(dataFrame[t])] for t in dataFrame ]
    colLengths = [ max([len(s) for s in colList]) for colList in listOfList ]
    
    listOfListPadded = [ [s.ljust(t[1]) for s in t[0]] for t in zip(listOfList,colLengths) ]
    
    return (listOfListPadded,colLengths,len(dataFrame.index)) 

#this function takes the output of 
def encode_frame(formattedFrame):
    return ([[encode(string) for string in column] for column in formattedFrame[0]],formattedFrame[1],formattedFrame[2])

def encode_list(dataList):
    colLength = max([len(s) for s in dataList])
    paddedList = [s.ljust(colLength) for s in dataList]
    return [encode(string) for string in paddedList]

"""
need to do some double checking on whether there can be issues with ordering change
"""

#uses the 'copy' module
#intended to take an encoded frame and have an atomic key integer added to every entry
def add_to_frame(encodedFrame,myKey):
    if sum(encodedFrame[1])*encodedFrame[2] != len(myKey):
        print(sum(encodedFrame[1])*encodedFrame[2],len(myKey))
        print(myKey)
        raise Exception("The supplied key is not the right size for the data.")
    
    #want a new encrypted frame and not to modify the old one
    toEncrypt = copy.deepcopy(encodedFrame)
    #one iteration over 
    atomicInd = 0 
    for colInd in range(len(encodedFrame[1])): 
        for rowInd in range(encodedFrame[2]): 
            for wordInd in range(encodedFrame[1][colInd]): 
                toEncrypt[0][colInd][rowInd][wordInd]+=myKey[atomicInd]['atom_key'] 
                atomicInd += 1
                
    return toEncrypt

#intended to take a cipher frame and decode before subtracting an atomic key integer from every entry
def sub_from_frame(cipherFrame,myKey,strip=False):
    if sum(cipherFrame.maxLength)*cipherFrame.rows != len(myKey):
        raise Exception("The supplied key is not the right size for the data.")    
    
    #want a new encrypted frame and not to modify the old one
    toDecrypt = copy.deepcopy(cipherFrame.ciphertext)
    #one iteration over 
    atomicInd = 0 
    for colInd in range(len(cipherFrame.maxLength)): 
        for rowInd in range(cipherFrame.rows): 
            for wordInd in range(cipherFrame.maxLength[colInd]): 
                toDecrypt[colInd][rowInd][wordInd]-=myKey[atomicInd]['atom_key'] 
                atomicInd += 1
            
            plainString = ''.join([chr(t) for t in toDecrypt[colInd][rowInd]])
                
            if strip == True:
                toDecrypt[colInd][rowInd] = plainString.strip()
            else:
                toDecrypt[colInd][rowInd] = plainString
            
    return toDecrypt

#uses the 'copy' module
#intended to take an encoded frame and have an atomic key integer added to every entry
def add_to_frame_gen(encodedFrame,myKeyGenerator):
    #!!!!!
    #need new error handling for length mismatch issues
   
    reverseMap = {}
    atomicInd = 0
    for colInd in range(len(encodedFrame[1])): 
        for rowInd in range(encodedFrame[2]): 
            for wordInd in range(encodedFrame[1][colInd]): 
                reverseMap[atomicInd] = {'colInd':colInd,'rowInd':rowInd,'wordInd':wordInd}
                atomicInd += 1
    
    #want a new encrypted frame and not to modify the old one - can we do better with memory here
    toEncrypt = copy.deepcopy(encodedFrame)
    #one iteration over 
    atomicInd = 0
    for atom in myKeyGenerator:
        #print(atomicInd,atom)
        colInd = reverseMap[atomicInd]['colInd']
        rowInd = reverseMap[atomicInd]['rowInd']
        wordInd = reverseMap[atomicInd]['wordInd']
        toEncrypt[0][colInd][rowInd][wordInd]+=atom['atom_key'] 
        atomicInd += 1
                
    return toEncrypt




# basic modules
import numpy as np
import pandas as pd
import json
from sqlalchemy import *
import urllib.parse

#a tapas of additional scientific computing 
from scipy.spatial import distance

#capnion submodules

from ..encoding import *
from ..ciphertext import *
from ..num_theory_toolbox import *
from ..db_toolbox import *
from ..polynomial import *

from .norm_cipher_num import NormCipherNum
from .paillier_num import PaillierFloat

class NormCipherQuant:
    def __init__(self, apiContext,cipherList,indexData=False,symbolic=False,
                 fromPlain=False,floatData=False,keyRange=2000,permLevel='standard'):
        
        if fromPlain == True:
            #in this scenario cipherIntegerList is actually a list of plaintext integers
            #register data
            myKeyLoc = apiContext.get('/state/?length='+str(len(cipherList))+'&range='
                                      +str(keyRange)+'&permLevel='+urllib.parse.quote(permLevel))[0]
            
            #determine key boundaries
            dataBoundary = [myKeyLoc['minId'],myKeyLoc['maxId']]
            
           
            #pull enc key and encrypt
            myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=False)
            keyData = []
            
            #use integer pads or float pads depending on data type
            if all(isinstance(x, int) for x in cipherList) and not floatData:
                for atom in myKeyGenerator:
                    keyData.append(atom['atom_key'])
                self.floatData = False
            else:
                for atom in myKeyGenerator:
                    keyData.append(atom['atom_key']+atom['atom_key_inv']/32767)
                #print(keyData)
                self.floatData = True
                
                
                
            #final addition step
            self.cipherList = [t[0]+t[1] for t in zip(cipherList,keyData)]
            self.indicesList = list(range(dataBoundary[0],dataBoundary[1]))
            
            #initialized with each symbolic entry referring directly to a 
            self.symbolic = ['x%d'%(i,) for i in self.indicesList]
            
        else:
            self.cipherList = cipherList
            self.floatData = floatData
            
            #generate ids if given only an idAnchor
            if isinstance(indexData,int):
                self.indicesList = list(range(indexData,indexData+len(cipherList)))
            #otherwise just use the given list of ids
            else:
                self.indicesList = indexData       
        
        self.length = len(cipherList)
        self.nums = [i for i in range(self.length)]
        self.apiContext = apiContext
        
        self.symbolic = symbolic

        
    def __len__(self):
        return self.length
        
    def __str__(self):
        return True
    
    def __hash__(self):
        return sum([(t**2 % 10000) for t in cipherList]) % 10000
    
    
    def __getslice__(self, start, stop):
        return NormCipherQuant(
            self.apiContext,
            self.cipherList[slice(start,stop)],
            self.indicesList[slice(start,stop)],
            floatData = self.floatData
        )

    #approximating a list of strings, so getting an item should give the normcipherstring
    def __getitem__(self, key):
        if isinstance(key, slice):
            return NormCipherQuant(
                self.apiContext,
                self.cipherList[key],
                self.indicesList[key],
                floatData = self.floatData
            )
        else:
            if type(self.nums[key]) != int:
                return self.nums[key]
            newNum = NormCipherNum(
                self.apiContext,
                self.cipherList[key],
                index = self.indicesList[key],
                floatData = self.floatData
            )
            self.nums[key] = newNum
            return newNum
    
    def __iter__(self):
        for i in range(len(self.cipherList)):
            yield self[slice(i,i+1)]

    def vert_slice(self,indexList):
        return NormCipherQuant(self.apiContext,
                               [self.cipherList[i] for i in indexList],
                               indexData=[self.indicesList[i] for i in indexList],
                               floatData = self.floatData
                              )
    
    def decrypt(self):
        if all(isinstance(x, int) for x in self.cipherList):
            
            decryptKeyDict = {t['id']:t['atom_key'] for t in decryption_key(
                self.apiContext,
                json.dumps(self.indicesList)
            )}
            decryptKey = [decryptKeyDict[i] for i in self.indicesList]
            #print(decryptKey)
            return [t[0]-t[1] for t in zip(self.cipherList,decryptKey)] 
        else:
            
            decryptKeyDict = {t['id']:(t['atom_key'],t['atom_key_inv']) for t in decryption_key(
                self.apiContext,
                json.dumps(self.indicesList)
            )}
            decryptKey = [decryptKeyDict[i] for i in self.indicesList]
            #print(decryptKey)
            return [t[0]-(t[1][0]+t[1][1]/32767) for t in zip(self.cipherList,decryptKey)] 
    
    def vert_merge(self, other):
        self.length = len(self.cipherList)+len(other.cipherList)
        return NormCipherQuant(
            self.apiContext,
            self.cipherList+other.cipherList,
            indexData = self.indicesList+other.indicesList,
            floatData = self.floatData
        )
    
    def mean(self):
        sumDecryptKey = ngram_checksum_key(
            self.apiContext,
            self.length,
            self.length,
            json.dumps(self.indicesList),
            isFloat=self.floatData
        )
        cipherSum = sum(self.cipherList)
        return (cipherSum - sumDecryptKey[0]['computed_range_sum'])/self.length

    def stdev(self):
        if self.length == 1:
            print("Must have length > 1 for standard deviation calculation")
            return "Length error"
        
        # get things we need for standard deviation calc
        mean = self.mean()
        polyString = "(x - {} )**2".format(mean)
        variables = ["x"]
        myIndices = [(i,) for i in self.indicesList]
        myCiphers = [(i,) for i in self.cipherList]
        
        # set up and solve the polynomial
        ans = full_polynomial_compute(self.apiContext,polyString,variables, myIndices, myCiphers, self.floatData,paillier=False,isSum=True,outPlain=True)
        
        std_dev = (ans/self.length) ** .5
        return std_dev
            
    def median(self):
        #need to change this to unknown polynomial
        variables = ['x']
        myIndices = [(i,) for i in self.indicesList]
        myCiphers = [(i,) for i in self.cipherList]
        # set up and solve the polynomial
       
        ans = full_polynomial_compute(self.apiContext,'random-median',variables, myIndices, myCiphers, self.floatData,paillier=False,isSum=False,outPlain=True)
        #print(ans)
        #sort the values
        index_val_tuples = []
        for i in range(self.length):
            index_val_tuples.append((i,ans[i]))
        
        sorted_list = sorted(index_val_tuples, key = lambda tup: tup[1])
        
        if len(sorted_list) % 2 == 0:
            #take the average of two in middle
            med_index1 = sorted_list[int(len(sorted_list)/2)][0]
            med_index2 = sorted_list[int(len(sorted_list)/2)-1][0]
            medNum = NormCipherNum(self.apiContext,self.cipherList[med_index1],index = self.indicesList[med_index1],floatData = self.floatData)
            medNum = medNum + NormCipherNum(self.apiContext,self.cipherList[med_index2],index = self.indicesList[med_index2],floatData = self.floatData)
            return medNum.decrypt()/2
            
        else:
            #return median
            med_index = med_index1 = sorted_list[len(sorted_list)//2][0]
            return NormCipherNum(self.apiContext,self.cipherList[med_index],index = self.indicesList[med_index],floatData = self.floatData)
        
        
    def cosine_similarity(self,other):
        if self.length != other.length:
            print("Cannot perform this operation on vectors of different lengths")
            return "Length Error"
        
        return self.dot_product(other) / (self.magnitude()*other.magnitude())
        
        
    def dot_product(self,other):
        if self.length != other.length:
            print("Cannot perform this operation on vectors of different lengths")
            return "Length Error"
        
        polyString = "x * y"
        variables = ['x', 'y']
        myIndices = [(self.indicesList[i],other.indicesList[i],) for i in range(self.length)]
        myCiphers = [(self.cipherList[i],other.cipherList[i],) for i in range(self.length)]
        # set up and solve the polynomial
        ans = full_polynomial_compute(self.apiContext,polyString,variables, myIndices, myCiphers, self.floatData,isSum=True,paillier=False,outPlain=True)
       
        return ans
        
    def coord_expon(self,n):
        polyString = "x ** %d" % (n,)
        variables = ['x']
        myIndices = [(i,) for i in self.indicesList]
        myCiphers = [(i,) for i in self.cipherList]
        # set up and solve the polynomial
        ans = full_polynomial_compute(self.apiContext,polyString,variables, myIndices, myCiphers, self.floatData,isSum=True,paillier=True,outPlain=True)
        return ans 
         
    def magnitude(self):
        polyString = "x ** 2"
        variables = ['x']
        myIndices = [(i,) for i in self.indicesList]
        myCiphers = [(i,) for i in self.cipherList]
        # set up and solve the polynomial
        ans = full_polynomial_compute(self.apiContext,polyString,variables, myIndices, myCiphers, self.floatData,isSum=True,paillier=True,outPlain=True)
        return ans ** .5
    
    def ciphertext(self,encodeList = False):
        cipherArray = []
        for i in range(len(self)):
            cipherArray.append(encode_ciphertext([self.cipherList[i]],encodeList))
        return cipherArray
    
    def list_of_ciphertext(self,encodeList = False):
        cipherArray = []
        for i in range(len(self)):
            cipherArray.append(encode_ciphertext([self.cipherList[i]],encodeList))
        return cipherArray
    
    def summation(self):
        sumDecryptKey = ngram_checksum_key(
            self.apiContext,
            self.length,
            self.length,
            json.dumps(self.indicesList),
            isFloat=self.floatData
        )
        cipherSum = sum(self.cipherList)
        #print(sumDecryptKey[0])
        return cipherSum - sumDecryptKey[0]['computed_range_sum']
    
    def __gt__(self,other):
        if isinstance(other,NormCipherNum):
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            
            return [int(num > 0) for num in plainResults]
        elif isinstance(other, (int, float, complex,np.int64,np.float64)):
            other = NormCipherNum(self.apiContext,other,fromPlain=True,keyRange=20)
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            
            return [int(num > 0) for num in plainResults]
        else:
            return False
        
    def __lt__(self,other):
        if isinstance(other,NormCipherNum):
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-comparison',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            print(plainResults)
            return [int(num < 0) for num in plainResults]
        elif isinstance(other, (int, float, complex,np.int64,np.float64)):
            other = NormCipherNum(self.apiContext,other,fromPlain=True,keyRange=20)
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-comparison',
                                              ['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            print(plainResults)
            return [int(num < 0) for num in plainResults]
        else:
            return False
        
        
    def __eq__(self,other):
        
        if isinstance(other,NormCipherNum):
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            
            numList = [int(round(num) == 0) for num in plainResults]
            return sum(numList) == len(numList)
        elif isinstance(other,NormCipherQuant):
            if len(other) != self.length:
                return False
            myIndices = [(self.indicesList[i],other.indicesList[i],) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipherList[i]] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            
            numList = [int(round(num) == 0) for num in plainResults]
            return sum(numList) == len(numList)
        
        elif isinstance(other,list):
            if len(other) != self.length:
                return False
            other = NormCipherQuant(self.apiContext,other,fromPlain=True,keyRange=20,floatData=self.floatData)
            myIndices = [(self.indicesList[i],other.indicesList[i],) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipherList[i]] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)            
            
            numList = [int(round(num) == 0) for num in plainResults]
            return sum(numList) == len(numList)
            
        elif isinstance(other, (int, float, complex,np.int64,np.float64)):
            other = NormCipherNum(self.apiContext,other,fromPlain=True,keyRange=20)
            myIndices = [(self.indicesList[i],other.index,) for i in range(len(self.indicesList))]
            cipherPair = [[self.cipherList[i],other.cipher] for i in range(len(self.cipherList))]

            plainResults = full_polynomial_compute(self.apiContext,'random-equality',['x','y'],myIndices,cipherPair,isFloat=self.floatData,paillier=False,outPlain=True)
            
            
            numList = [int(round(num) == 0) for num in plainResults]
            return sum(numList) == len(numList)
        else:
            return False
    



# basic modules
import numpy as np
import pandas as pd
import json
from sqlalchemy import *

#a tapas of additional scientific computing 
from scipy.spatial import distance

#capnion submodules

from ..encoding import *
from ..ciphertext import *
from ..num_theory_toolbox import *
from ..db_toolbox import *
from ..polynomial import *

from .norm_cipher_string import *

class NormCipherList:

    def __init__(self,apiContext,cipherListOfList,indexData=False,fromPlain=False,seedString=False,keyRange=2000,permLevel='standard'):
        
        if fromPlain:
            importData = import_and_encrypt_list(cipherListOfList,apiContext,seedString,keyRange)
            self.cipherListOfList = importData[0]
            self.indicesListOfList = importData[1]
            
            
        else:
            #get list of group elements
            #the rawest version of the underlying data is a list of list of integers
            self.cipherListOfList = cipherListOfList
             #assumes entries the same length
            

            #index data may be an integer, indicating an idAnchor and continuous indices
            #or it may be a an explicit list of indices
            if isinstance(indexData,int):
                self.indicesListOfList = [] 
                i=indexData
                for cipherList in self.cipherListOfList:
                    self.indicesListOfList.append(list(range(i,i+len(cipherList))))
                    i+=len(cipherList)
            else:
                self.indicesListOfList = indexData
                
        self.colMaxChars = len(self.cipherListOfList[0])        
        self.apiContext = apiContext
        self.words = [i for i in range(len(self.cipherListOfList))]
        self.length = len(self.cipherListOfList)
        self.helper = 0
        
    def __len__(self):
        return len(self.cipherListOfList)#//self.maxLength

    def __getslice__(self, start, stop):
        return NormCipherList(
            self.apiContext,
            self.cipherListOfList[slice(start,stop)],
            indexData = self.indicesListOfList[slice(start,stop)]
        )

    #approximating a list of strings, so getting an item should give the normcipherstring
    def __getitem__(self, key):
        
        if isinstance(key, slice):
            return NormCipherList(
                self.apiContext,
                self.cipherListOfList[key],
                indexData = self.indicesListOfList[key]
            )
        else:
            if isinstance(self.words[key], int):
                ncs = NormCipherString(
                    self.apiContext,
                    self.cipherListOfList[key],
                    indexData = self.indicesListOfList[key]
                    )
                self.words[key] = ncs
                return ncs
            else:
                return self.words[key]
        
    def __iter__(self):
        for i in range(len(self.cipherListOfList)):
            yield self[slice(i,i+1)]
            
    def pad(self,charsToAdd):
        encryptPadding = NormCipherFrame(self.apiContext,pd.DataFrame([[' '*charsToAdd]*len(self.cipherListOfList)]).transpose())[0]
        return NormCipherList(
            self.apiContext,
            [
                cipherListTuple[0] + cipherListTuple[1] 
                for cipherListTuple in zip(self.cipherListOfList,encryptPadding.cipherListOfList)
            ],
            indexData = [
                indicesListTuple[0] + indicesListTuple[1] 
                for indicesListTuple in zip(self.indicesListOfList,encryptPadding.indicesListOfList)
            ]
        )
        
    def vert_merge(self, other):
        if type(other) == NormCipherList:
            return NormCipherList(
                self.apiContext,
                self.cipherListOfList+other.cipherListOfList,
                indexData = self.indicesListOfList+other.indicesListOfList
            )
        else:
            return NormCipherList(
                self.apiContext,
                self.cipherListOfList+[other.cipherList],
                indexData = self.indicesListOfList+[other.indicesList]
            )

    #the linking key endpoint needs to be updated to take into account the possibility of... 
    #...a more complicated list of indices
    def char_equal_mx(self):
        #using polynomial endpoint
        myCiphers = [[cipher] for word in self.cipherListOfList for cipher in word]
        myIndices = [[index] for word in self.indicesListOfList for index in word]
        
        
        plainResults = full_polynomial_compute(self.apiContext,'random-sort',['x'],myIndices,
                                               myCiphers,False,paillier=False,outPlain=True)
        #print(plainResults)
        charEqMx = []
        
        charsPerCol = len(self.cipherListOfList * self.colMaxChars)
        for i in range(len(plainResults)):
            for j in range(len(plainResults)):
                if round(plainResults[i]) == round(plainResults[j]):
                    charEqMx.append(1)
                else:
                    charEqMx.append(0)
                    
        outputArray = np.array(charEqMx)
        outputArray.shape = (charsPerCol,charsPerCol)
        #print(outputArray)
        return outputArray
    
    def ngram_hashes(self,n):
        ngramDecryptKey = [t['computed_range_sum'] for t in ngram_checksum_key(
            self.apiContext,
            n,
            self.colMaxChars,
            json.dumps(flatten_list(self.indicesListOfList))
        )]
        #print(ngramDecryptKey)
        i = 0
        ngramHashes = []
        for cellList in self.cipherListOfList:
            cellHashes = []
            for wordIndex in range(self.colMaxChars-n+1):
                cellHashes.append(sum(cellList[slice(wordIndex,wordIndex+n)])-ngramDecryptKey[i])
                i+=1
            ngramHashes.append(cellHashes)
        return ngramHashes
    
    def ngram_distance_matrix(self,n):
        hashes = self.ngram_hashes(n)
        outputArray = np.zeros((len(hashes),len(hashes)))
        for i in range(len(hashes)):
            for j in range(len(hashes)):
                outputArray[i,j] = distance.cosine(hashes[i],hashes[j])
        return outputArray
    
    def list_of_ciphertext(self,encodeList=False):
        return [encode_ciphertext(self[t].cipherList,encodeList) for t in range(len(self))]
    
    #this function homomorphically computes a simple hash function returning the sum of the encoded values
    #this presumes all entries in the list are the same length
    def checksum(self):
        checksumDecryptKey = ngram_checksum_key( 
            self.apiContext,
            self.colMaxChars,
            self.colMaxChars,
            json.dumps(flatten_list(self.indicesListOfList))
        )
        #print(self.colMaxChars)
        #print(checksumDecryptKey)
        checksumList = [
            t[0]-t[1]['computed_range_sum'] 
            for t in zip([sum(u) for u in self.cipherListOfList],checksumDecryptKey)
        ]
        return checksumList

    #this is for finding
    #this will require the API
    def search(self,queryString,**kwargs):
        wordLength = len(self[0])
        if isinstance(queryString,str):
            # check for lengths and pad accordingly
            
            if len(queryString) > wordLength:
                return False
            while len(queryString) < wordLength:
                queryString += ' '
                
            queryString = NormCipherString(self.apiContext,queryString,keyRange=200)
            
        
        # make sure we have a normCipherString now
        if not isinstance(queryString,NormCipherString):
            return False
        
        tempList = self.vert_merge(queryString)
        
        tempMx = tempList.char_equal_mx()
        #print(len(tempMx))
        #tempList.helper = AnalyticsHelper(self.apiContext,temp_mx,tempList.colMaxChars)

        indexMatches = []
        #return tempMx
        # check equivalence

        for i in range(wordLength):
            if i == 0:
                for j in range(self.length):
                    if tempMx[-1*wordLength][j*wordLength] == 1:
                        indexMatches.append(j)
                #print(indexMatches)
            else:
                for match in indexMatches:
                    if tempMx[-1*wordLength+i][match*wordLength+i] == 1:
                        pass
                    else:
                        indexMatches.remove(match)
        
        if indexMatches == []:
            return False
        return indexMatches #indices of occurrence
    
    #first attempt at computing a levenshtein distance for entity detection purposes
    def levenshtein(self):

        adjMx = self.char_equal_mx()
        word_length = int(len(adjMx[0])/len(self))
        levenMx = np.zeros((len(self),len(self)))
        for i in range(len(self)):
            for j in range(len(self)):
                word_dist = np.zeros((len(self[i])+1,len(self[j])+1))
                for u in range(1,len(self[i])+1):
                    word_dist[u][0] = u
                for v in range(1,len(self[j])+1):
                    word_dist[0][v] = v
                for ii in range(1,len(self[i])+1):
                    for jj in range(1,len(self[j])+1):
                        subPenalty = 1 if adjMx[(i*word_length+ii-1)][(j*word_length+jj-1)] == 0 else 0
                        word_dist[ii][jj] = min(word_dist[ii-1][jj]+1,word_dist[ii][jj-1]+1,word_dist[ii-1][jj-1]+subPenalty)
                levenMx[i][j] = word_dist[len(self[i])][len(self[j])]

            return levenMx
    
    
    
    def custom_equality(self,func):
        
        def getAndRun():
            
            if self.helper == 0:
                temp_mx = self.char_equal_mx()
                self.helper = AnalyticsHelper(self.apiContext,temp_mx,self.colMaxChars)
            analyticsList = []
            for key in range(len(self.cipherListOfList)):
                analyticsList.append(AnalyticsCipherString(self.apiContext,
                                                           self.cipherListOfList[key],
                                                           self.helper,
                                                           self.indicesListOfList[key],
                                                           key*self.helper.offset))
            return func(analyticsList)
                
                
        return getAndRun()
    
    def decrypt(self):
        decryptKeyDict = {t['id']:t['atom_key'] for t in decryption_key(
            self.apiContext,
            json.dumps(flatten_list(self.indicesListOfList))
        )}
        decryptKey = [decryptKeyDict[i] for i in flatten_list(self.indicesListOfList)]
        i=0
        plain = []
        for encryptedWord in self.cipherListOfList:
            plain.append(''.join([chr(t[0]-t[1]) for t in zip(encryptedWord,decryptKey[i:(i+len(encryptedWord))])]))
            i+=len(encryptedWord)
        return plain
    
    def generate_matches(self,other):
        lengthDiff = self.colMaxChars - other.colMaxChars
        if lengthDiff > 0:
            compareOne = self.checksum()
            compareTwo = other.pad(lengthDiff).checksum()
        elif lengthDiff < 0:
            compareOne = self.pad(lengthDiff).checksum()
            compareTwo = other.checksum()
        else:
            compareOne = self.checksum()
            compareTwo = other.checksum()            
            
        matches = []
        hasMatchOne = []
        hasMatchTwo = []
        for i in range(len(compareOne)):
            currentHash = compareOne[i]
            for j in range(len(compareTwo)):
                if compareOne[i]==compareTwo[j]:
                    matches.append((i,j))
                    hasMatchOne.append(i)
                    hasMatchTwo.append(j)
                    
        checkNCL = NormCipherList(
            self.apiContext,
            [self.cipherListOfList[i] for i in hasMatchOne],
            indexData = [self.indicesListOfList[i] for i in hasMatchOne]
        ).vert_merge(
            NormCipherList(
                self.apiContext,
                [other.cipherListOfList[i] for i in hasMatchTwo],
                indexData = [other.indicesListOfList[i] for i in hasMatchTwo]
            )
        )


        matchCheck = checkNCL.char_equal_mx()
        
        wordLength = self.colMaxChars
        compareLength = len(matches)
        newMatches = []
        for i in range(compareLength):
            spotCheck = np.all(np.diagonal(matchCheck[
                (wordLength*i):(wordLength*i+wordLength),
                (compareLength*wordLength+wordLength*i):(compareLength*wordLength+wordLength*i+wordLength)
            ])==1)
            if spotCheck:
                newMatches.append(matches[i])
        return newMatches
      
      
    # this function returns a list of NormCipherLists grouped by value (hash value currently)
    def group_by(self):

        listOfGroupedNCL = []
        indicesAlreadyAdded = []

        if self.helper == 0:
            temp_mx = self.char_equal_mx()
            self.helper = AnalyticsHelper(self.apiContext,temp_mx,self.colMaxChars)

        checksumList = self.checksum()
        for i in range(len(self)):
            value = checksumList[i]
            newCipherListOfList = []
            newIndicesListOfList = []
            if i not in indicesAlreadyAdded:
                indicesAlreadyAdded.append(i)
                for j in range(len(self)):
                    ncs1 = self[i]
                    ncs2 = self[j]
                    if self.helper.checkEquality(ncs1.indicesList,ncs2.indicesList,i*self.colMaxChars,j*self.colMaxChars):
                        newCipherListOfList.append(self[j].cipherList)
                        newIndicesListOfList.append(self[j].indicesList)
                        indicesAlreadyAdded.append(j)
                newNCL = NormCipherList(self.apiContext, newCipherListOfList, newIndicesListOfList)
                listOfGroupedNCL.append(newNCL)

        return listOfGroupedNCL
    
    
    
    
def import_and_encrypt_list(myPlaintext,apiContext,desiredPerms=False,seedString=False,keyRange=32766,permLevel='standard'):
    
    

    columnData = list(myPlaintext)
    myEncodedList = encode_list(columnData)
    myLen = len(myEncodedList)*len(myEncodedList[0])

    #compute length of required encryption key


    #register key data of required length

    if isinstance(seedString,str):
        myKeyLoc = apiContext.get('/statehash/?length=%d&seedString=%s'%(myLen,seedString,))[0]
        #print('basic state')
    else:
        myKeyLoc = apiContext.get('/state/?length='+str(myLen)+'&range='+str(keyRange)+'&permLevel='+permLevel)[0]


    #create encryption key for current user

    dataBoundary = [myKeyLoc['minId'],myKeyLoc['maxId']]


    #pull enc key and encrypt

    myKeyGenerator = encryption_key(apiContext,dataBoundary,htmlDebug=False,seedString=seedString)
    #print(myKeyGenerator)
    sortedKeyGenerator = {}
    for atom in myKeyGenerator:
        sortedKeyGenerator[str(atom['id'])] = atom

    currentIndex = myKeyLoc['minId']
    cipherListOfList = []
    indexListOfList = []
    for i in range(len(myEncodedList)):
        cipherList = []
        indexList = []
        for j in range(len(myEncodedList[0])):
            cipherList.append(myEncodedList[i][j]+int(sortedKeyGenerator[str(currentIndex)]['atom_key']))
            indexList.append(currentIndex)
            currentIndex += 1
        cipherListOfList.append(cipherList)
        indexListOfList.append(indexList)        



    return (
        cipherListOfList,
        indexListOfList,
    )